from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help='Show the available environments')
@click.argument('envs', required=False, nargs=-1)
@click.option('--ascii', 'force_ascii', is_flag=True, help='Whether or not to only use ASCII characters')
@click.option('--json', 'as_json', is_flag=True, help='Whether or not to output in JSON format')
@click.option('--internal', '-i', is_flag=True, help='Show internal environments')
@click.option('--hide-titles', is_flag=True, hidden=True)
@click.pass_obj
def show(
    app: Application,
    *,
    envs: tuple[str, ...],
    force_ascii: bool,
    as_json: bool,
    internal: bool,
    hide_titles: bool,
):
    """Show the available environments."""
    app.ensure_environment_plugin_dependencies()

    from hatch.config.constants import AppEnvVars

    if as_json:
        import json

        contextual_config = {}
        for environments in (app.project.config.envs, app.project.config.internal_envs):
            for env_name, config in environments.items():
                environment = app.project.get_environment(env_name)
                new_config = contextual_config[env_name] = dict(config)

                env_vars = dict(environment.env_vars)
                env_vars.pop(AppEnvVars.ENV_ACTIVE)
                if env_vars:
                    new_config['env-vars'] = env_vars

                num_dependencies = len(config.get('dependencies', []))
                dependencies = environment.environment_dependencies[:num_dependencies]
                if dependencies:
                    new_config['dependencies'] = dependencies

                extra_dependencies = environment.environment_dependencies[num_dependencies:]
                if extra_dependencies:
                    new_config['extra-dependencies'] = extra_dependencies

                if environment.pre_install_commands:
                    new_config['pre-install-commands'] = list(
                        environment.resolve_commands(environment.pre_install_commands)
                    )

                if environment.post_install_commands:
                    new_config['post-install-commands'] = list(
                        environment.resolve_commands(environment.post_install_commands)
                    )

                if environment.scripts:
                    new_config['scripts'] = {
                        script: list(environment.resolve_commands([script])) for script in environment.scripts
                    }

        app.display(json.dumps(contextual_config, separators=(',', ':')))
        return

    from packaging.requirements import InvalidRequirement, Requirement

    from hatchling.metadata.utils import get_normalized_dependency, normalize_project_name

    if internal:
        target_standalone_envs = app.project.config.internal_envs
        target_matrices = app.project.config.internal_matrices
    else:
        target_standalone_envs = app.project.config.envs
        target_matrices = app.project.config.matrices

    for env_name in envs:
        if env_name not in target_standalone_envs and env_name not in target_matrices:
            app.abort(f'Environment `{env_name}` is not defined by project config')

    env_names = set(envs)

    matrix_columns: dict[str, dict[int, str]] = {
        'Name': {},
        'Type': {},
        'Envs': {},
        'Features': {},
        'Dependencies': {},
        'Environment variables': {},
        'Scripts': {},
        'Description': {},
    }
    matrix_envs = set()
    for i, (matrix_name, matrix_data) in enumerate(target_matrices.items()):
        matrix_envs.update(matrix_data['envs'])

        if env_names and matrix_name not in env_names:
            continue

        config = matrix_data['config']
        matrix_columns['Name'][i] = matrix_name
        matrix_columns['Type'][i] = config['type']
        matrix_columns['Envs'][i] = '\n'.join(matrix_data['envs'])

        if config.get('features'):
            if app.project.metadata.hatch.metadata.allow_ambiguous_features:
                matrix_columns['Features'][i] = '\n'.join(sorted(set(config['features'])))
            else:
                matrix_columns['Features'][i] = '\n'.join(
                    sorted({normalize_project_name(f) for f in config['features']})
                )

        dependencies = []
        if config.get('dependencies'):
            dependencies.extend(config['dependencies'])
        if config.get('extra-dependencies'):
            dependencies.extend(config['extra-dependencies'])
        if dependencies:
            normalized_dependencies = set()
            for dependency in dependencies:
                try:
                    req = Requirement(dependency)
                except InvalidRequirement:
                    normalized_dependencies.add(dependency)
                else:
                    normalized_dependencies.add(get_normalized_dependency(req))

            matrix_columns['Dependencies'][i] = '\n'.join(sorted(normalized_dependencies))

        if config.get('env-vars'):
            matrix_columns['Environment variables'][i] = '\n'.join(
                '='.join(item) for item in sorted(config['env-vars'].items())
            )

        if config.get('scripts'):
            matrix_columns['Scripts'][i] = '\n'.join(
                sorted(script for script in config['scripts'] if app.verbose or not script.startswith('_'))
            )

        if config.get('description'):
            matrix_columns['Description'][i] = config['description'].strip()

    standalone_columns: dict[str, dict[int, str]] = {
        'Name': {},
        'Type': {},
        'Features': {},
        'Dependencies': {},
        'Environment variables': {},
        'Scripts': {},
        'Description': {},
    }
    standalone_envs = (
        (env_name, config)
        for env_name, config in target_standalone_envs.items()
        if env_names or env_name not in matrix_envs
    )
    for i, (env_name, config) in enumerate(standalone_envs):
        if env_names and env_name not in env_names:
            continue

        environment = app.project.get_environment(env_name)

        standalone_columns['Name'][i] = env_name
        standalone_columns['Type'][i] = config['type']

        if environment.features:
            standalone_columns['Features'][i] = '\n'.join(environment.features)

        if environment.environment_dependencies_complex:
            standalone_columns['Dependencies'][i] = '\n'.join(
                sorted({get_normalized_dependency(d) for d in environment.environment_dependencies_complex})
            )

        env_vars = dict(environment.env_vars)
        env_vars.pop(AppEnvVars.ENV_ACTIVE)
        if env_vars:
            standalone_columns['Environment variables'][i] = '\n'.join(
                '='.join(item) for item in sorted(env_vars.items())
            )

        if environment.scripts:
            standalone_columns['Scripts'][i] = '\n'.join(
                sorted(script for script in environment.scripts if app.verbose or not script.startswith('_'))
            )

        if environment.description:
            standalone_columns['Description'][i] = environment.description.strip()

    column_options = {}
    for title in matrix_columns:
        if title != 'Description':
            column_options[title] = {'no_wrap': True}

    app.display_table(
        '' if hide_titles else 'Standalone',
        standalone_columns,
        show_lines=True,
        column_options=column_options,
        force_ascii=force_ascii,
    )
    app.display_table(
        '' if hide_titles else 'Matrices',
        matrix_columns,
        show_lines=True,
        column_options=column_options,
        force_ascii=force_ascii,
    )



from __future__ import annotations

import json
import os
import shutil
import subprocess
import sys
from tempfile import TemporaryDirectory

RUNNER: dict = {}


def main() -> int:
    project_root: str = RUNNER['project_root']
    output_dir: str = RUNNER['output_dir']
    hook: str = RUNNER['hook']
    kwargs: dict[str, str] = RUNNER['kwargs']
    backend: str = RUNNER['backend']
    backend_path: str = RUNNER['backend_path']
    hook_caller_script: str = RUNNER['hook_caller_script']

    with TemporaryDirectory() as d:
        temp_dir = os.path.realpath(d)
        control_dir = os.path.join(temp_dir, 'control')
        os.mkdir(control_dir)
        input_file = os.path.join(control_dir, 'input.json')
        output_file = os.path.join(control_dir, 'output.json')

        env_vars = dict(os.environ)
        env_vars['_PYPROJECT_HOOKS_BUILD_BACKEND'] = backend
        if backend_path:
            env_vars['_PYPROJECT_HOOKS_BACKEND_PATH'] = backend_path

        if 'work_dir' in kwargs:
            work_dir = os.path.join(temp_dir, 'work')
            os.mkdir(work_dir)
            kwargs[kwargs.pop('work_dir')] = work_dir
        else:
            work_dir = ''

        with open(input_file, 'w', encoding='utf-8') as f:
            f.write(json.dumps({'kwargs': kwargs}))

        script_path = os.path.join(temp_dir, 'script.py')
        with open(script_path, 'w', encoding='utf-8') as f:
            f.write(hook_caller_script)

        process = subprocess.run(
            [sys.executable, script_path, hook, str(control_dir)],
            cwd=project_root,
            env=env_vars,
            check=False,
        )
        if process.returncode:
            return process.returncode

        with open(output_file, encoding='utf-8') as f:
            output = json.loads(f.read())

        if output.get('no_backend', False):
            sys.stderr.write(f'{output["traceback"]}\n{output["backend_error"]}\n')
            return 1

        if output.get('unsupported', False):
            sys.stderr.write(output['traceback'])
            return 1

        if output.get('hook_missing', False):
            sys.stderr.write(f'Build backend API `{backend}` is missing hook: {output["missing_hook_name"]}\n')
            return 1

        shutil.move(output_file, output_dir)
        if work_dir:
            shutil.move(work_dir, output_dir)

    return 0


if __name__ == '__main__':
    code = main()
    sys.stderr.flush()
    os._exit(code)



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help='Locate Python binaries')
@click.argument('name')
@click.option('-p', '--parent', is_flag=True, help='Show the parent directory of the Python binary')
@click.option('--dir', '-d', 'directory', help='The directory in which distributions reside')
@click.pass_obj
def find(app: Application, *, name: str, parent: bool, directory: str | None):
    """Locate Python binaries."""
    manager = app.get_python_manager(directory)
    installed = manager.get_installed()
    if name not in installed:
        app.abort(f'Distribution not installed: {name}')

    dist = installed[name]
    app.display(str(dist.python_path.parent if parent else dist.python_path))



"""Backport of @functools.singledispatchmethod to Python <3.7.

Adapted from https://github.com/ikalnytskyi/singledispatchmethod
removing 2.7 specific code.
"""

import functools
import typing

if hasattr(functools, "singledispatchmethod"):  # pragma: nocover
    singledispatchmethod = functools.singledispatchmethod  # type: ignore
else:
    update_wrapper = functools.update_wrapper
    singledispatch = functools.singledispatch

    # The type: ignore below is to avoid mypy erroring due to a
    # "already defined" singledispatchmethod, oddly this does not
    # happen when using `if sys.version_info >= (3, 8)`

    class singledispatchmethod(object):  # type: ignore
        """Single-dispatch generic method descriptor.

        TODO: Figure out how to type this:

        `mypy --strict` returns errors like the following for all decorated methods:
        "Untyped decorator makes function "send" untyped."

        But this is not a normal function-base decorator, it's a class and it
        doesn't have a __call__ method. When decorating the "base" method
        __init__ is called, but of course its return type is None.
        """

        def __init__(self, func: typing.Callable[..., typing.Any]) -> None:
            if not callable(func) and not hasattr(func, "__get__"):
                raise TypeError("{!r} is not callable or a descriptor".format(func))

            self.dispatcher = singledispatch(func)
            self.func = func

        def register(
            self,
            cls: typing.Callable[..., typing.Any],
            method: typing.Optional[typing.Callable[..., typing.Any]] = None,
        ) -> typing.Callable[..., typing.Any]:
            """Register a method on a class for a particular type.

            Note in Python <= 3.6 this methods cannot infer the type from the
            argument's type annotation, users *must* supply it manually on
            decoration, i.e.

            @my_method.register(TypeToDispatch)
            def _(self, arg: TypeToDispatch) -> None:
                ...

            Versus in Python 3.7+:

            @my_method.register
            def _(self, arg: TypeToDispatch) -> None:
                ...

            """
            # mypy wants method to be non-optional, but it is required to be
            # for decoration to work correctly in our case.
            # https://github.com/python/cpython/blob/3.8/Lib/functools.py#L887-L920
            # is not type annotated either.
            return self.dispatcher.register(cls, func=method)  # type: ignore

        def __get__(
            self, obj: typing.Any, cls: typing.Callable[[typing.Any], typing.Any]
        ) -> typing.Callable[..., typing.Any]:
            def _method(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
                method = self.dispatcher.dispatch(args[0].__class__)  # type: typing.Any
                return method.__get__(obj, cls)(*args, **kwargs)

            # The type: ignore below is due to `_method` being given a strict
            # "Callable[[VarArg(Any), KwArg(Any)], Any]" which causes a
            # 'has no attribute "__isabstractmethod__" error'
            # felt safe enough to ignore
            _method.__isabstractmethod__ = self.__isabstractmethod__  # type: ignore
            _method.register = self.register  # type: ignore
            update_wrapper(_method, self.func)
            return _method

        @property
        def __isabstractmethod__(self) -> typing.Any:
            return getattr(self.func, "__isabstractmethod__", False)



class AppEnvVars:
    ENV = 'HATCH_ENV'
    ENV_ACTIVE = 'HATCH_ENV_ACTIVE'
    ENV_OPTION_PREFIX = 'HATCH_ENV_TYPE_'
    QUIET = 'HATCH_QUIET'
    VERBOSE = 'HATCH_VERBOSE'
    INTERACTIVE = 'HATCH_INTERACTIVE'
    PYTHON = 'HATCH_PYTHON'
    # https://no-color.org
    NO_COLOR = 'NO_COLOR'
    FORCE_COLOR = 'FORCE_COLOR'


class ConfigEnvVars:
    PROJECT = 'HATCH_PROJECT'
    DATA = 'HATCH_DATA_DIR'
    CACHE = 'HATCH_CACHE_DIR'
    CONFIG = 'HATCH_CONFIG'


class PublishEnvVars:
    USER = 'HATCH_INDEX_USER'
    AUTH = 'HATCH_INDEX_AUTH'
    REPO = 'HATCH_INDEX_REPO'
    CA_CERT = 'HATCH_INDEX_CA_CERT'
    CLIENT_CERT = 'HATCH_INDEX_CLIENT_CERT'
    CLIENT_KEY = 'HATCH_INDEX_CLIENT_KEY'
    PUBLISHER = 'HATCH_PUBLISHER'
    OPTIONS = 'HATCH_PUBLISHER_OPTIONS'


class PythonEnvVars:
    CUSTOM_SOURCE_PREFIX = 'HATCH_PYTHON_CUSTOM_SOURCE_'
    CUSTOM_PATH_PREFIX = 'HATCH_PYTHON_CUSTOM_PATH_'
    CUSTOM_VERSION_PREFIX = 'HATCH_PYTHON_CUSTOM_VERSION_'


class VersionEnvVars:
    VALIDATE_BUMP = 'HATCH_VERSION_VALIDATE_BUMP'



import os
import sys

FILE = os.path.abspath(__file__)
HERE = os.path.dirname(FILE)
ENV_VAR_PREFIX = '_HATCHLING_PORT_ADD_'


def _apply_env_vars(kwargs):
    from ast import literal_eval

    for key, value in os.environ.items():
        if key.startswith(ENV_VAR_PREFIX):
            kwargs[key.replace(ENV_VAR_PREFIX, '', 1).lower()] = literal_eval(value)


def _parse_dependencies(dependency_definition):
    dependencies = []
    for line in dependency_definition.splitlines():
        dependency = line.split(' #', 1)[0].strip()
        if dependency:
            dependencies.append(dependency)

    return dependencies


def _collapse_data(output, data):
    import tomli_w

    expected_output = new_output = tomli_w.dumps(data)
    new_output = new_output.replace('    ', '')
    new_output = new_output.replace('[\n', '[')
    new_output = new_output.replace('",\n]', '"]')

    return output.replace(expected_output, new_output, 1)


def _parse_setup_cfg(kwargs):
    from configparser import ConfigParser

    setup_cfg_file = os.path.join(HERE, 'setup.cfg')
    if not os.path.isfile(setup_cfg_file):
        return

    setup_cfg = ConfigParser()
    setup_cfg.read(setup_cfg_file, encoding='utf-8')

    if not setup_cfg.has_section('metadata'):
        return

    metadata = setup_cfg['metadata']

    if 'name' in metadata and 'name' not in kwargs:
        kwargs['name'] = metadata['name']

    if 'description' in metadata and 'description' not in kwargs:
        kwargs['description'] = metadata['description']

    if 'license' in metadata and 'license' not in kwargs:
        kwargs['license'] = metadata['license']

    if 'author' in metadata and 'author' not in kwargs:
        kwargs['author'] = metadata['author']

    if 'author_email' in metadata and 'author_email' not in kwargs:
        kwargs['author_email'] = metadata['author_email']

    if 'keywords' in metadata and 'keywords' not in kwargs:
        keywords = metadata['keywords'].strip().splitlines()
        kwargs['keywords'] = keywords if len(keywords) > 1 else keywords[0]

    if 'classifiers' in metadata and 'classifiers' not in kwargs:
        kwargs['classifiers'] = metadata['classifiers'].strip().splitlines()

    if 'url' in metadata and 'url' not in kwargs:
        kwargs['url'] = metadata['url']

    if 'download_url' in metadata and 'download_url' not in kwargs:
        kwargs['download_url'] = metadata['download_url']

    if 'project_urls' in metadata and 'project_urls' not in kwargs:
        kwargs['project_urls'] = dict(
            url.replace(' = ', '=', 1).split('=') for url in metadata['project_urls'].strip().splitlines()
        )

    if setup_cfg.has_section('options'):
        options = setup_cfg['options']

        if 'python_requires' in options and 'python_requires' not in kwargs:
            kwargs['python_requires'] = options['python_requires']

        if 'install_requires' in options and 'install_requires' not in kwargs:
            kwargs['install_requires'] = _parse_dependencies(options['install_requires'])

        if 'packages' in options and 'packages' not in kwargs:
            packages = []
            for package_spec in options['packages'].strip().splitlines():
                package = package_spec.replace('find:', '', 1).replace('find_namespace:', '', 1).strip()
                if package:
                    packages.append(package)

            if packages:
                kwargs['packages'] = packages

        if 'package_dir' in options and 'package_dir' not in kwargs:
            kwargs['package_dir'] = dict(
                d.replace(' = ', '=', 1).split('=') for d in options['package_dir'].strip().splitlines()
            )

    if setup_cfg.has_section('options.extras_require') and 'extras_require' not in kwargs:
        kwargs['extras_require'] = {
            feature: _parse_dependencies(dependencies)
            for feature, dependencies in setup_cfg['options.extras_require'].items()
        }

    if setup_cfg.has_section('options.entry_points') and 'entry_points' not in kwargs:
        kwargs['entry_points'] = {
            entry_point: definitions.strip().splitlines()
            for entry_point, definitions in setup_cfg['options.entry_points'].items()
        }


def setup(**kwargs):
    import itertools
    import re

    import tomli_w

    project_metadata = {}
    hatch_metadata = {}
    tool_metadata = {'hatch': hatch_metadata}
    project_data = {
        'build-system': {'requires': ['hatchling'], 'build-backend': 'hatchling.build'},
        'project': project_metadata,
        'tool': tool_metadata,
    }

    _parse_setup_cfg(kwargs)
    _apply_env_vars(kwargs)

    name = kwargs['name']
    project_name = name.replace('_', '-')
    packages = sorted(kwargs.get('packages') or [name.replace('-', '_')])
    package_name = package_path = package_source = packages[0].split('.')[0].lower().strip()

    project_metadata['name'] = project_name

    project_metadata['dynamic'] = ['version']

    if 'description' in kwargs:
        project_metadata['description'] = kwargs['description']

    for readme_file in ('README.md', 'README.rst', 'README.txt'):
        if os.path.isfile(os.path.join(HERE, readme_file)):
            project_metadata['readme'] = readme_file
            break

    project_metadata['license'] = kwargs.get('license', '')

    if 'python_requires' in kwargs:
        project_metadata['requires-python'] = kwargs['python_requires']

    for collaborator in ('author', 'maintainer'):
        collaborators = []
        collaborator_names = []
        collaborator_emails = []
        if collaborator in kwargs:
            collaborator_names.extend(
                collaborator_name.strip() for collaborator_name in kwargs[collaborator].split(',')
            )
        if f'{collaborator}_email' in kwargs:
            collaborator_emails.extend(
                collaborator_email.strip() for collaborator_email in kwargs[f'{collaborator}_email'].split(',')
            )

        for collaborator_name, collaborator_email in itertools.zip_longest(collaborator_names, collaborator_emails):
            data = {}
            if collaborator_name is not None:
                data['name'] = collaborator_name
            if collaborator_email is not None:
                data['email'] = collaborator_email
            if data:
                collaborators.append(data)

        if collaborators:
            project_metadata[f'{collaborator}s'] = collaborators

    if 'keywords' in kwargs:
        keywords = kwargs['keywords']
        if isinstance(keywords, str):
            keywords = keywords.replace(',', ' ').split()
        project_metadata['keywords'] = sorted(keywords)

    if 'classifiers' in kwargs:
        project_metadata['classifiers'] = sorted(kwargs['classifiers'])
        fixed_indices = []
        final_index = 0
        for i, classifier in enumerate(project_metadata['classifiers']):
            if classifier.startswith('Programming Language :: Python :: '):
                final_index = i
                for python_version in ('3.10', '3.11', '3.12'):
                    if classifier.endswith(python_version):
                        fixed_indices.append(i)
                        break
        for i, index in enumerate(fixed_indices):
            project_metadata['classifiers'].insert(final_index, project_metadata['classifiers'].pop(index - i))

    if 'install_requires' in kwargs:
        project_metadata['dependencies'] = sorted(
            [entry.strip() for entry in kwargs['install_requires']]
            if isinstance(kwargs['install_requires'], (list, tuple))
            else _parse_dependencies(kwargs['install_requires']),
            key=lambda d: d.lower(),
        )

    if 'extras_require' in kwargs:
        project_metadata['optional-dependencies'] = {
            group: sorted(
                [entry.strip() for entry in dependencies]
                if isinstance(dependencies, (list, tuple))
                else _parse_dependencies(dependencies),
                key=lambda d: d.lower(),
            )
            for group, dependencies in sorted(kwargs['extras_require'].items())
        }

    if 'entry_points' in kwargs and isinstance(kwargs['entry_points'], dict):
        entry_points = {}
        for entry_point, raw_definitions in kwargs['entry_points'].items():
            definitions = [raw_definitions] if isinstance(raw_definitions, str) else raw_definitions
            definitions = dict(sorted(d.replace(' ', '').split('=', 1) for d in definitions))

            if entry_point == 'console_scripts':
                project_metadata['scripts'] = definitions
            elif entry_point == 'gui_scripts':
                project_metadata['gui-scripts'] = definitions
            else:
                entry_points[entry_point] = definitions
        if entry_points:
            project_metadata['entry-points'] = dict(sorted(entry_points.items()))

    urls = {}
    if 'url' in kwargs:
        urls['Homepage'] = kwargs['url']
    if 'download_url' in kwargs:
        urls['Download'] = kwargs['download_url']
    if 'project_urls' in kwargs:
        urls.update(kwargs['project_urls'])
    if urls:
        project_metadata['urls'] = dict(sorted(urls.items()))

    build_targets = {}
    build_data = {}

    if 'use_scm_version' in kwargs:
        project_data['build-system']['requires'].append('hatch-vcs')
        hatch_metadata['version'] = {'source': 'vcs'}
        build_data['hooks'] = {'vcs': {'version-file': f'{package_path}/_version.py'}}
    else:
        hatch_metadata['version'] = {'path': f'{package_path}/__init__.py'}

    build_data['targets'] = build_targets

    if '' in kwargs.get('package_dir', {}):
        package_source = kwargs['package_dir']['']

        package = (kwargs.get('packages') or [package_name])[0]
        package_path = f'{package_source}/{package}'

        if package_path != f'src/{package_name}':
            build_targets.setdefault('wheel', {})['packages'] = [package_path]

    if kwargs.get('data_files', []):
        shared_data = {}
        for shared_directory, relative_paths in kwargs['data_files']:
            relative_files = {}
            for relative_path in relative_paths:
                relative_directory, filename = os.path.split(relative_path)
                relative_files.setdefault(relative_directory, []).append(filename)

            for relative_directory, files in sorted(relative_files.items()):
                if not os.path.isdir(relative_directory) or set(os.listdir(relative_directory)) != set(files):
                    for filename in sorted(files):
                        local_path = os.path.join(relative_directory, filename).replace('\\', '/')
                        shared_data[local_path] = f'{shared_directory}/{filename}'
                else:
                    shared_data[relative_directory] = shared_directory

        build_targets.setdefault('wheel', {})['shared-data'] = shared_data

    build_targets['sdist'] = {
        'include': [
            f'/{package_source}',
        ]
    }

    hatch_metadata['build'] = build_data

    output = tomli_w.dumps(project_data)
    output = _collapse_data(output, {'requires': project_data['build-system']['requires']})
    output = _collapse_data(output, {'dynamic': project_data['project']['dynamic']})

    project_file = os.path.join(HERE, 'pyproject.toml')
    if os.path.isfile(project_file):
        with open(project_file, encoding='utf-8') as f:
            current_contents = f.read()

        for section in ('build-system', 'project'):
            for pattern in (rf'^\[{section}].*?(?=^\[)', rf'^\[{section}].*'):
                current_contents = re.sub(pattern, '', current_contents, flags=re.MULTILINE | re.DOTALL)

        output += f'\n{current_contents}'

    with open(project_file, 'w', encoding='utf-8') as f:
        f.write(output)


if __name__ == 'setuptools':
    __this_shim = sys.modules.pop('setuptools')
    __current_directory = sys.path.pop(0)

    import setuptools as __real_setuptools

    sys.path.insert(0, __current_directory)
    sys.modules['setuptools'] = __this_shim

    def __getattr__(name):
        return getattr(__real_setuptools, name)

    del __this_shim
    del __current_directory


def migrate(root, setuptools_options, sys_paths):
    import shutil
    import subprocess
    from tempfile import TemporaryDirectory

    with TemporaryDirectory() as temp_dir:
        repo_dir = os.path.join(os.path.realpath(temp_dir), 'repo')
        shutil.copytree(root, repo_dir, ignore=shutil.ignore_patterns('.git', '.tox'), copy_function=shutil.copy)
        shutil.copy(FILE, os.path.join(repo_dir, 'setuptools.py'))
        setup_py = os.path.join(repo_dir, 'setup.py')

        if not os.path.isfile(setup_py):
            # Synthesize a small setup.py file since there is none
            with open(setup_py, 'w', encoding='utf-8') as f:
                f.write('from setuptools import setup\nsetup()\n')

        env = dict(os.environ)
        for arg in setuptools_options:
            key, value = arg.split('=', 1)
            env[f'{ENV_VAR_PREFIX}{key}'] = value

        # When PYTHONSAFEPATH is non-empty, the current directory is not added automatically
        python_paths = [repo_dir]
        python_paths.extend(p for p in sys_paths if p)
        if python_path := env.get('PYTHONPATH', ''):
            python_paths.append(python_path)

        env['PYTHONPATH'] = os.pathsep.join(python_paths)

        subprocess.run([sys.executable, setup_py], env=env, cwd=repo_dir, check=True)

        old_project_file = os.path.join(root, 'pyproject.toml')
        new_project_file = os.path.join(repo_dir, 'pyproject.toml')
        shutil.copyfile(new_project_file, old_project_file)
